@@ -1,6 +1,6 @@
 """
 MedUX - A Free/OpenSource Electronic Medical Record
-Copyright (C) 2017-2021 Christian González
+Copyright (C) 2017-2022 Christian González
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU Affero General Public License as published
@@ -15,38 +15,28 @@ GNU Affero General Public License for more details.
 You should have received a copy of the GNU Affero General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 """
-from typing import Union
-
-from django.core.exceptions import ValidationError
-from django.core.validators import RegexValidator
-from django.utils.translation import ugettext_lazy as _
-from django.db import models
-from django.core import exceptions
-from django.utils.datetime_safe import datetime
-
-import base64
-from uuid import uuid4
-
-from environ import ImproperlyConfigured
-
-from medux.core.validators import (
-    fhir_server_allowed_references,
-    CodeValidator,
-    IdValidator,
-    OidValidator,
-)
-
-__author__ = "Christian González <christian.gonzalez@nerdocs.at>"
 
 # TODO: Fields could be done better, see http://build.fhir.org/datatypes.html
 # These types should be implemented very carefully, with all the validators
 # and custom behaviours in place.
-# Especially the ReferenceField is used very often and should be implemented
-# very well.
+__author__ = "Christian González <christian.gonzalez@nerdocs.at>"
 
+import base64
+from datetime import timedelta
+from decimal import Decimal
+from enum import Enum
+from uuid import uuid4
 
-from django import forms
+from composite_field import CompositeField
 from django.conf import settings
+from django.core.exceptions import ImproperlyConfigured
+from django.core.validators import RegexValidator, MinValueValidator
+from django.db import models
+from django.utils.functional import cached_property
+from django.utils.translation import gettext_lazy as _
+from enumfields import EnumField
+
+from .validators import CodeValidator, OidValidator, IdValidator
 
 
 class OptionalTimeDateTimeField(models.DateTimeField):
@@ -59,9 +49,9 @@ class OptionalTimeDateTimeField(models.DateTimeField):
         field = None
         try:
             if settings.MEDUX["BIRTHDAY_FORMAT"] == "Date":
-                field = forms.DateField
+                field = "forms.DateField"
             elif settings.MEDUX["BIRTHDAY_FORMAT"] == "DateTime":
-                field = forms.DateTimeField
+                field = "forms.DateTimeField"
             else:
                 raise ImproperlyConfigured(
                     f"MEDUX['BIRTHDAY_FORMAT'] settings variable contains wrong value '{settings.MEDUX['BIRTHDAY_FORMAT']}'. Please set it to 'Date' or 'DateTime'."
@@ -79,13 +69,6 @@ class OptionalTimeDateTimeField(models.DateTimeField):
         )
 
 
-# ===================== Primitive types =====================
-# Primitive types are those with only a value,
-# and no additional elements as children.
-# see examples: http://build.fhir.org/datatypes-examples.html#primitives
-# ===========================================================
-
-
 class Base64TextField(models.TextField):
     """A stream of bytes, base64 encoded"""
 
@@ -102,11 +85,6 @@ class Base64TextField(models.TextField):
     # TODO: read only yet
 
 
-# StringFields are implemented either as Textfields, or as Charfields,
-# depending on how many chars (>255) are possible. Some databases do not
-# implement more than 255 chars in a string.
-
-
 class StringListField(models.TextField):
     """Represents a list of strings.
 
@@ -119,12 +97,12 @@ class StringListField(models.TextField):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
-    def to_python(self, value):
-        """Normalize data to a list of strings."""
-        # Return an empty list if no input was given.
-        if not value:
-            return []
-        return value.split("\n")
+    # def to_python(self, value):
+    #     """Normalize data to a list of strings."""
+    #     # Return an empty list if no input was given.
+    #     if not value:
+    #         return []
+    #     return value.split("\n")
 
 
 class InstantField(models.DateTimeField):
@@ -153,64 +131,6 @@ class InstantField(models.DateTimeField):
         return name, path, args, kwargs
 
 
-# ===================== Complex types =====================
-# In XML, these types are represented as XML Elements with child elements
-# with the name of the defined elements of the type.
-# The name of the element is defined where the type is used.
-# In JSON, the data type is represented by an object with properties named
-# the same as the XML elements. Since the JSON representation is almost
-# exactly the same, only the first example has an additional explicit
-# JSON representation.
-#
-# Complex data types may be "profiled".
-# A Structure Definition or type "constraint" makes a set of rules about
-# which elements SHALL have values and what the possible values are.
-# ===========================================================
-
-
-class ManyReferenceField(models.ManyToManyField):
-    def __init__(self, to: str, **kwargs):
-
-        self.allowed_references = to.split("|")
-        for ref in self.allowed_references:
-            if ref not in fhir_server_allowed_references.split("|"):
-                raise exceptions.FieldError(
-                    _("'{}' is not allowed as reference in {}".format(ref, "<FIXME>"))
-                )
-
-        # No matter what this field should (dynamically) refer to,
-        # always make sure the Foreign key is bound to a "Reference" object
-        kwargs["to"] = "Reference"
-        super().__init__(**kwargs)
-
-    def deconstruct(self):
-        name, path, args, kwargs = super().deconstruct()
-        kwargs["to"] = "|".join(self.allowed_references)
-        return name, path, args, kwargs
-
-
-class CodeableConceptField(models.ForeignKey):
-    def __init__(self, value_set: str, *args, **kwargs):
-        assert type(value_set) == str
-
-        self.allowed_codings = value_set.split("|")
-
-        # always refer to the model "CodeableConcept", no matter what is given
-        kwargs["to"] = "CodeableConcept"
-
-        # FIXME: is SET_NULL ok everywhere?
-        kwargs["null"] = True
-        kwargs["on_delete"] = models.SET_NULL
-        super().__init__(*args, **kwargs)
-
-    def deconstruct(self):
-        name, path, args, kwargs = super().deconstruct()
-        del kwargs["null"]
-        kwargs["value_set"] = "|".join(self.allowed_codings)
-        return name, path, args, kwargs
-
-
-# url is represented by Django's URLField
 class UriField(models.URLField):
     """A Uniform Resource Identifier Reference.
 
@@ -253,22 +173,8 @@ class CanonicalField(UriField):
     """
 
 
-# class PositiveIntField(models.PositiveIntegerField):
-#     description = _("Positive integer >= 1")
-#
-#     def formfield(self, **kwargs):
-#         defaults = {"min_value": 1}
-#         defaults.update(kwargs)
-#         return super().formfield(**defaults)
-
-
-# class UnsignedIntField(models.PositiveIntegerField):
-#     description = _("Positive integer >= 0")
-#
-#     # this is the Django implementation of PositiveIntegerField
-
-
-class CodeField(models.CharField):
+# FIXME: better use TextChoices here?
+class CodeField(EnumField):
     """Represents a field with a "code" that is defined elsewhere.
 
     This field must not contain arbitrary strings, but only the ones that are defined in the given
@@ -276,40 +182,56 @@ class CodeField(models.CharField):
 
     Technically, a code is restricted to a string which has at least one
     character and no leading or trailing whitespace, and where there is no
-    whitespace other than single spaces in the contents"""
+    whitespace other than single spaces in the contents.
+
+    A code is determined by 4 elements:
+
+    * system - the system the code lives in, e.g. "http://loinc.org"
+    * version - the version this code has in the system, e.g. "1.27"
+    * code - the code itself, e.g. "55423-8", or "active"
+    * display - the human readable representation, e.g. "The prescription is active"
+    """
 
     # TODO: This could be a ForeignKey, pointing to a FHIR ValueSet class!
     description = _('A "code" that is defined elsewhere')
     default_validators = [CodeValidator]
-    default_error_messages = {
-        "not_multiple": _('There are no multiple values allowed in "{key}".'),
-    }
+    # default_error_messages = {
+    #     "not_multiple": _('There are no multiple values allowed in "{key}".'),
+    # }
 
     def __init__(
-        self, terminology_binding: str = None, multi: bool = False, *args, **kwargs
+        self,
+        terminology_binding: type(Enum) = None,
+        # system: str = None,
+        # version: str = None,
+        **kwargs,
     ):
 
-        assert terminology_binding is None or type(terminology_binding) == str
+        if terminology_binding is not None and not issubclass(
+            terminology_binding, Enum
+        ):
+            raise TypeError(
+                f"CodeField: 'technology_binding' parameter must be a Enum, but is '{type(terminology_binding)}'"
+            )
 
+        # TODO: do something with system/version etc.
+        # self.system = system
+        # self.version = version
         # The ValueSet this code is defined in
         self.terminology_binding = terminology_binding
-        # are multiple values allowed?
-        self.multi = multi
-
-        # TODO: we set this to an arbitrary 64 char string as max.
-        # Could be more specific
-        kwargs["max_length"] = 64
-
-        # value_set = "Patient"  # FIXME: this is just a demo
-        # does not work: ValueSet.objects.name(name=terminology_binding).first() # etc
 
         # if not value_set:
         #     raise exceptions.ValidationError(
         #         "Terminology binding {} not found.".format(terminology_binding)
         #     )
-        choices = {}
-        kwargs["choices"] = choices
-        super().__init__(*args, **kwargs)
+        # if terminology_binding:
+        #     kwargs["choices"] = [(tag.value, tag.name) for tag in terminology_binding]
+        # if not terminology_binding:
+        #     terminology_binding = Enum
+        kwargs["max_length"] = 32
+        if "enum" in kwargs:
+            print("found!", kwargs)
+        super().__init__(terminology_binding, **kwargs)
 
     def validate(self, value, model_instance):
         super().validate(value, model_instance)
@@ -325,9 +247,8 @@ class CodeField(models.CharField):
 
     def deconstruct(self):
         name, path, args, kwargs = super().deconstruct()
-        del kwargs["max_length"]
+        del kwargs["enum"]
         kwargs["terminology_binding"] = self.terminology_binding
-        kwargs["multi"] = self.multi
         return name, path, args, kwargs
 
 
@@ -354,7 +275,7 @@ class IdField(models.CharField):
 
     def __init__(self, *args, **kwargs):
         kwargs["max_length"] = 64
-        kwargs["default"] = uuid4()
+        kwargs["default"] = uuid4
         kwargs["validators"] = [IdValidator]
         super().__init__(*args, **kwargs)
 
@@ -374,14 +295,238 @@ class MarkdownField(models.TextField):
     # TODO: implement Validator: \s*(\S|\s)*
 
 
+class NarrativeStatus(Enum):
+    GENERATED = "generated"
+    EXTENSIONS = "extensions"
+    ADDITIONAL = "additional"
+    EMPTY = "empty"
+
+
+class PeriodField(CompositeField):
+    description = "A Field representing a period in time, with optional start or end."
+
+    # Constraints: One of [start|end] must be given.
+
+    start = models.DateTimeField(blank=True, null=True)
+    end = models.DateTimeField(blank=True, null=True)
+
+    # def __repr__(self):
+    #     return Period(start=self.start, end=self.end)
+
+    def timedelta(self) -> timedelta:
+        return self.end - self.start
+
+
+class UnitField(models.CharField):
+    description = "A field which stores units."
+
+    def __init__(self, **kwargs):
+        kwargs.setdefault("max_length", 10)
+        super().__init__(**kwargs)
+
+
+class QuantityField(CompositeField):
+    COMPARATORS = (
+        ("<", _("<")),
+        ("<=", _("<=")),
+        (">=", _(">=")),
+        (">", _(">")),
+    )
+
+    description = (
+        "A field which stores value and (optionally) unit of a quantity., e.g. '250 ml'"
+    )
+
+    value = models.DecimalField(max_length=None, decimal_places=None)
+    unit = UnitField(blank=True)
+    system = models.CharField(max_length=255, blank=True)
+    # code = CodeField(terminology_binding=...)  # The (FHIR) preferred system is UCUM
+    comparator = models.CharField(max_length=2, blank=True, choices=COMPARATORS)
+
+    def __init__(
+        self,
+        max_digits: int | None,
+        decimal_places: int | None,
+        verbose_name: str = None,
+        blank: bool = False,
+        default: Decimal = None,
+        validators=None,
+    ):
+        super().__init__()
+        self.verbose_name = verbose_name
+        value = self["value"]
+        value.blank = blank
+        value.decimal_places = decimal_places
+        value.max_digits = max_digits
+        value.null = blank
+        value.blank = blank
+        self._validators = validators
+        if default is not None:
+            self.value.default = default.real
+
+    def __str__(self):
+        return f"{self.value} {self.unit}"
+
+    # def clean(self, value, model_instance):
+    #     """
+    #     We need to run validation against ``Quantity`` instance.
+    #     """
+    #     output = self.to_python(value)
+    #     self.validate(value, model_instance)
+    #     self.run_validators(value)
+    #     return output
+
+    @cached_property
+    def validators(self):
+        # Default ``DecimalValidator`` doesn't work with ``Quantity`` instances.
+        # FIXME use correct validators
+        return super().validators + [self._validators]
+
+    def formfield(self, **kwargs):
+        defaults = {
+            "form_class": "core.QuantityFormField",
+            "decimal_places": self.value.decimal_places,
+        }
+        defaults.update(kwargs)
+        return super().formfield(**defaults)
+
+
+class RatioField(CompositeField):
+    """A relationship between two Quantity values.
+
+    Examples where a Ratio is used are:
+
+    * titers (e.g. 1:150)
+    * concentration ratios where the denominator is significant (e.g. 5mg/dl)
+    * observed frequencies (e.g. 2 repetitions/8 hr)
+    """
+
+    numerator = QuantityField(
+        max_digits=1, decimal_places=1, blank=True
+    )  # 1=placeholder
+    denominator = QuantityField(
+        max_digits=1, decimal_places=1, validators=[MinValueValidator(0)]
+    )  # 1=placeholder
+
+    def __init__(
+        self,
+        num_max_digits: int,
+        num_decimal_places: int,
+        denom_max_digits: int,
+        denom_decimal_places: int,
+        verbose_name: str = None,
+        blank: bool = False,
+        validators=None,
+    ):
+        super().__init__()
+        self.verbose_name = verbose_name
+
+        numerator = self["numerator"]
+        numerator.max_digits = num_max_digits
+        numerator.decimal_places = num_decimal_places
+
+        denominator = self["denominator"]
+        denominator.max_digits = denom_max_digits
+        denominator.decimal_places = denom_decimal_places
+
+        numerator.blank = blank
+        denominator.blank = blank
+        denominator.null = False  # TODO: make sure null works here
+        self._validators = validators
+
+    def __str__(self):
+        return f"{self.numerator}/{self.denominator}"
+
+
+# ======================== Unused Fields from FHIR ========================
+
+
+# ===================== Primitive types =====================
+# Primitive types are those with only a value,
+# and no additional elements as children.
+# see examples: http://build.fhir.org/datatypes-examples.html#primitives
+# ===========================================================
+
+
+# StringFields are implemented either as Textfields, or as Charfields,
+# depending on how many chars (>255) are possible. Some databases do not
+# implement more than 255 chars in a string.
+
+
+# ===================== Complex types =====================
+# In XML, these types are represented as XML Elements with child elements
+# with the name of the defined elements of the type.
+# The name of the element is defined where the type is used.
+# In JSON, the data type is represented by an object with properties named
+# the same as the XML elements. Since the JSON representation is almost
+# exactly the same, only the first example has an additional explicit
+# JSON representation.
+#
+# Complex data types may be "profiled".
+# A Structure Definition or type "constraint" makes a set of rules about
+# which elements SHALL have values and what the possible values are.
+# ===========================================================
+
+
+# class ManyReferenceField(models.ManyToManyField):
+#     def __init__(self, to: str, **kwargs):
+#
+#         self.allowed_references = to.split("|")
+#         for ref in self.allowed_references:
+#             if ref not in fhir_server_allowed_references.split("|"):
+#                 raise exceptions.FieldError(
+#                     _("'{}' is not allowed as reference in {}".format(ref, "<FIXME>"))
+#                 )
+#
+#         # No matter what this field should (dynamically) refer to,
+#         # always make sure the Foreign key is bound to a "Reference" object
+#         kwargs["to"] = "Reference"
+#         super().__init__(**kwargs)
+#
+#     def deconstruct(self):
+#         name, path, args, kwargs = super().deconstruct()
+#         kwargs["to"] = "|".join(self.allowed_references)
+#         return name, path, args, kwargs
+
+
+# class CodeableConceptField(models.ForeignKey):
+#     def __init__(self, value_set: str, *args, **kwargs):
+#         assert type(value_set) == str
+#
+#         self.allowed_codings = value_set.split("|")
+#
+#         # always refer to the model "CodeableConcept", no matter what is given
+#         kwargs["to"] = "CodeableConcept"
+#
+#         # FIXME: is SET_NULL ok everywhere?
+#         kwargs["null"] = True
+#         kwargs["on_delete"] = models.SET_NULL
+#         super().__init__(*args, **kwargs)
+#
+#     def deconstruct(self):
+#         name, path, args, kwargs = super().deconstruct()
+#         del kwargs["null"]
+#         kwargs["value_set"] = "|".join(self.allowed_codings)
+#         return name, path, args, kwargs
+
+
+# class PositiveIntField(models.PositiveIntegerField):
+#     description = _("Positive integer >= 1")
+#
+#     def formfield(self, **kwargs):
+#         defaults = {"min_value": 1}
+#         defaults.update(kwargs)
+#         return super().formfield(**defaults)
+
+
+# class UnsignedIntField(models.PositiveIntegerField):
+#     description = _("Positive integer >= 0")
+#
+#     # this is the Django implementation of PositiveIntegerField
+
+
 # http://hl7.org/fhir/narrative-status
 # FIXME: implement/import as ValueSet
-NARRATIVE_STATUS = (
-    ("generated", _("Generated")),
-    ("extensions", _("Extensions")),
-    ("additional", _("Additional")),
-    ("empty", _("Empty")),
-)
 
 
 # ======================== Special data types ========================
@@ -456,58 +601,16 @@ NARRATIVE_STATUS = (
 #     div = models.TextField(null=False, blank=False)
 
 
-class Period:
-    """A period of Time, with optional start/end"""
-
-    def __init__(self, start: datetime = None, end: datetime = None):
-        self.start = start
-        self.end = end
-
-    def __len__(self):
-        return 2 * 25 + 1
-
-
-class PeriodField(models.CharField):
-    """A Field representing a period in time, with optional start or end.
-
-    One of [start|end] must be given."""
-
-    # TODO: make a date(time) widget
-
-    description = "A period of Time, with optional start/end"
-
-    def __init__(self, *args, **kwargs):
-        kwargs["max_length"] = 2 * 25 + 1
-        super().__init__(*args, **kwargs)
-
-    def deconstruct(self) -> tuple:
-        name, path, args, kwargs = super().deconstruct()
-        del kwargs["max_length"]
-        return name, path, args, kwargs
-
-    def _parse(self, value: str) -> Period:
-        start, end = value.split(";")
-        if start:  # datetime
-            start = datetime.fromisoformat(start)
-        if end:
-            end = datetime.fromisoformat(end)
-        return Period(start, end)
-
-    def to_python(self, value) -> Union[Period, None]:
-        if isinstance(value, Period):
-            return value
-
-        if value is None:
-            return value
-
-        if isinstance(value, str):
-            return self._parse(value)
-
-        raise ValidationError(f"Could not parse Period value: {value}")
-
-    def get_prep_value(self, value: Period) -> str:
-        """serializes a Period value into a db saveable string"""
-        if isinstance(value, Period):
-            start = str(value.start.isoformat()) if value.start else ""
-            end = str(value.end.isoformat()) if value.end else ""
-            return ";".join([start, end])
+# def validate_quantity_value(value):
+#     """
+#     Valid value for Quantity are:
+#       - Single numeric value
+#       - Quantity instances
+#       - Pairs of numeric value and unit. unit can't be None.
+#     """
+#     if isinstance(value, (list, tuple)) and (len(value) != 2 or value[1] is None):
+#         raise ValidationError(
+#             f"Invalid value for QuantityField: {value}.",
+#             code="invalid",
+#         )
+#
