@@ -1,8 +1,10 @@
-from .user import User
 from .common import *
-from .fhir import *
+
+# from .fhir import *
 from .patient import *
 from .datapacks import *
 from .geo import *
 from .medication import *
 from .organizations import *
+from .devices import *
+from .observations import *
