@@ -4,17 +4,23 @@ from django.contrib.auth import views
 from django.contrib.auth.views import LoginView
 
 from .views import HomeView, PatientFileView
+from .views.medication_detail import CreatePrescriptionView
 from .views.patient_file import NewPatientView, PatientListView, SettingsView
 
 app_name = "core"
 
 root_urlpatterns = [
     path("", HomeView.as_view(), name="home"),
-    path("settings/", SettingsView.as_view(), name="settings"),
+    path("preferences/", SettingsView.as_view(), name="preferences"),
     path("patient/", PatientListView.as_view(), name="patient_list"),
     path("patient/file/<pk>", PatientFileView.as_view(), name="patient_file"),
     path("patient/add/", NewPatientView.as_view(), name="patient_new"),
     path("accounts/", include("django.contrib.auth.urls")),
+    path(
+        "medication/prescriptions/add",
+        CreatePrescriptionView.as_view(),
+        name="create_prescription",
+    ),
 ]
 
 # Django login accounts views:
