@@ -30,11 +30,13 @@ class MeduxBaseMixin(PermissionRequiredMixin):
 class HomeView(MeduxBaseMixin, TemplateView):
     template_name = "home.html"
     # FIXME find a correct permission for home, and use MeduxBaseMixin
-    permission_required = ()
     extra_context = {
         "viewmodes": iter(IViewMode),
     }
 
+    def has_permission(self):
+        return self.request.user.is_authenticated
+
 
 class PatientFileView(MeduxBaseMixin, DetailView):
     model = Patient
